import pandas as pd
from pyadlml.dataset.activities import correct_activities
from pyadlml.dataset.devices import correct_devices
from pyadlml.dataset.obj import Data
from pyadlml.dataset import START_TIME, END_TIME, DEVICE, VAL, TIME

DATA_NAME = 'devices.csv'
DEV_MAP_NAME = 'device_mapping.csv'
ACT_NAME = 'activities_subject_%s.csv'

def _read_activities(path_to_file):
    activities = pd.read_csv(path_to_file)
    activities[START_TIME] = pd.to_datetime(activities[START_TIME])
    activities[END_TIME] = pd.to_datetime(activities[END_TIME])
    return activities

def _read_devices(path_to_dev_file, path_to_mapping):
    devices = pd.read_csv(path_to_dev_file)
    dev_map = pd.read_csv(path_to_mapping, index_col='id')\
                .to_dict()[DEVICE] 
    devices[DEVICE] = devices[DEVICE].map(dev_map)
    devices[VAL] = devices[VAL].astype(bool)
    devices[TIME] = pd.to_datetime(devices[TIME])
    devices = devices.reset_index(drop=True)
    return devices

def load(folder_path, subject):
    """
    """
    df_dev = _read_devices(folder_path + '/' + DATA_NAME,
                            folder_path + '/' + DEV_MAP_NAME)
    df_act = _read_activities(folder_path + '/' + ACT_NAME%(subject))

    # correct possible overlaps in activities
    df_act, cor_lst = correct_activities(df_act)
    
    # correct possible duplicates for representation 2    
    df_dev = correct_devices(df_dev)
    data = Data(df_act, df_dev)
    data.correction_activities = cor_lst
    return data



import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.pyplot as plt
import matplotlib
from scipy.stats.kde import gaussian_kde
from scipy.stats import norm
from sklearn.preprocessing import minmax_scale
import copy
from matplotlib.colors import LogNorm


def hm_key_NN(hm, num):
    """ a nearest neighbor mapping using keys
    Parameters
    ----------
    hm : dict
        contains key value mappings
    num : int
        the key value for which the nearest key should be chosen
    Returns
    -------
    res : mapping
    """
    nearest_neighbor = None
    min_dist = np.inf
    for key in hm:
       diff = abs(key-num) 
       if diff < min_dist:
           min_dist = diff
           nearest_neighbor = key
    return hm[nearest_neighbor]
           

def func_formatter_log(x, pos):
    x = np.exp(x)
    if x-60 < 0:
        return "{:.0f}s".format(x)
    elif x-3600 < 0:
        return "{:.0f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.0f}h".format(x/3600)
    else:
        return "{:.0f}t".format(x/86400)

def func_formatter_sec(x, pos):
    if x-60 < 0:
        return "{:.1f}s".format(x)
    elif x-3600 < 0:
        return "{:.1f}m".format(x/60)
    elif x-86400 < 0:
        return "{:.1f}h".format(x/3600)
    else:
        return "{:.1f}t".format(x/86400)

def func_formatter_min(x, pos):
    if x-60 < 0:
        return "{:.0f}m".format(x)
    elif x-3600 < 0:
        return "{:.0f}h".format(x/60)
    else:
        return "{:.0f}t".format(x/3600)

def heatmap_contingency(vals, acts, devs, cbarlabel, title, valfmt, figsize, \
     z_scale=None, numbers=True):
    # cut off the name of the device for every second label to make it more readable
    for i in range(0,len(devs)):
        if i % 2 == 0:
            tmp = devs[i][:-4]
            devs[i] = tmp + ' Off'
        else:
            devs[i] = 'On'

    if z_scale == 'log':
        log = True
    else:
        log = False

    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap(vals, acts, devs, ax=ax, log=log, cbarlabel=cbarlabel)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt=valfmt)        

    # create grid for heatmap into every pair
    tcks = np.arange((vals.shape[1])/2)*2 + 1.5
    ax.set_xticks(tcks, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=2)
    ax.tick_params(which="minor", bottom=False, left=False) 

    ax.set_title(title)
    fig.tight_layout()
    plt.show()

def heatmap(data, row_labels, col_labels, log=False, ax=None,
            cbar_kw={}, cbarlabel="", **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    row_labels
        A list or array of length N with the labels for the rows.
    col_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    cbar_kw
        A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
    cbarlabel
        The label for the colorbar.  Optional.
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if not ax:
        ax = plt.gca()

    # Plot the heatmap
    if log:
        cbarlabel = "log " + cbarlabel
        
        #set colorlabel for bad colors
        cmap = copy.copy(matplotlib.cm.get_cmap("viridis"))
        cmap.set_bad(color = 'lightgrey', alpha = .8)
        
        # round to the next highest digit for display purposes
        vmax = 10**int(np.ceil(np.log10(data.max())))
        vmin = 1
        
        im = ax.imshow(data, cmap=cmap, norm=LogNorm(vmin=vmin, vmax=vmax),**kwargs)
    else:
        im = ax.imshow(data, **kwargs)

    # Create colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size='5%', pad=0.05)
    cbar = cax.figure.colorbar(im, cax, **cbar_kw) 
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")
    #cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Turn spines off and create white grid.
    #for edge, spine in ax.spines.items():
    #    spine.set_visible(False)

    return im, cbar

def heatmap_square(data, row_labels, col_labels, log=False, ax=None,
            cbar_kw={}, cbarlabel="", grid=True, **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    row_labels
        A list or array of length N with the labels for the rows.
    col_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    cbar_kw
        A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
    cbarlabel
        The label for the colorbar.  Optional.
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if not ax:
        ax = plt.gca()

    # Plot the heatmap
    if log:
        cbarlabel = "log " + cbarlabel
        
        #set colorlabel for bad colors
        cmap = copy.copy(matplotlib.cm.get_cmap("viridis"))
        cmap.set_bad(color = 'lightgrey', alpha = .8)
        
        # round to the next highest digit for display purposes
        vmax = 10**int(np.ceil(np.log10(data.max())))
        vmin = 1
        
        im = ax.imshow(data, cmap=cmap, norm=LogNorm(vmin=vmin, vmax=vmax),**kwargs)
    else:
        im = ax.imshow(data, **kwargs)

        
    # Create colorbar
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size='5%', pad=0.05)
    cbar = cax.figure.colorbar(im, cax, **cbar_kw) # debug
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")
    if grid:
        # Turn spines off and create white grid.
        ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
        ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
        ax.grid(which="minor", color="w", linestyle='-', linewidth=1)
        ax.tick_params(which="minor", bottom=False, left=False)

    return im, cbar


def annotate_heatmap(im, data=None, valfmt="{}",
                     textcolors=("black", "white"),
                     threshold=None, log=False, **textkw):
    """
    A function to annotate a heatmap.

    Parameters
    ----------
    im
        The AxesImage to be labeled.
    data
        Data used to annotate.  If None, the image's data is used.  Optional.
    valfmt
        The format of the annotations inside the heatmap.  This should either
        use the string format method, e.g. "$ {x:.2f}", or be a
        `matplotlib.ticker.Formatter`.  Optional.
    textcolors
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.  Optional.
    threshold
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.  Optional.
    **kwargs
        All other arguments are forwarded to each call to `text` used to create
        the text labels.
    """
    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max())/2.

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if log and not isinstance(valfmt, matplotlib.ticker.FuncFormatter):
        def func_formatter_tmp(x,p):
            with np.errstate(divide='ignore'):
                x = np.log10(x)
                if np.isinf(x):
                    #return "-$\infty$"
                    return ""
                else:
                    return "{:.2f}".format(x)
        
        format_func = lambda x, p: func_formatter_tmp(x, p)
        valfmt = matplotlib.ticker.FuncFormatter(format_func)
    elif isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)
        
    diverging_lst = [ 'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']
    
    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            # if the content is masked for e.g log values set the color to black
            if isinstance(im.norm(data[i,j]), np.ma.core.MaskedConstant):
                kw.update(color='black')
                
            # if list is diverging high values should have white and values
            # near zero should have black values
            elif im.cmap.name in diverging_lst:
                if 1 - im.norm(data[i, j]) >= threshold or im.norm(data[i, j]) >= threshold:
                    color = 'white'
                else:
                    color = 'black'
                kw.update(color=color)
            else:
                kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts

def ridgeline(data, overlap=0, fill=True, labels=None, n_points=150, dist_scale=0.05):
    """
    Creates a standard ridgeline plot.

    data, list of lists.
    overlap, overlap between distributions. 1 max overlap, 0 no overlap.
    fill, matplotlib color to fill the distributions.
    n_points, number of points to evaluate each distribution function.
    labels, values to place on the y axis to describe the distributions.
    """
    if overlap > 1 or overlap < 0:
        raise ValueError('overlap must be in [0 1]')
    xx = np.linspace(np.min(np.concatenate(data)),
                     np.max(np.concatenate(data)), n_points)
    
    curves = []
    ys = []

    for i, d in enumerate(data):
        pdf = gaussian_kde(d)
        y = i*(1.0-overlap)
        ys.append(y)
        curve = pdf(xx)
        curve = minmax_scale(curve)*dist_scale
        if fill:
            plt.fill_between(xx, np.ones(n_points)*y, 
                             curve+y, zorder=len(data)-i+1, color=fill)
        plt.plot(xx, curve+y, c='k', zorder=len(data)-i+1)
        #break
    if labels:
        plt.yticks(ys, labels)

def savefig(fig, file_path):
    """ saves figure to folder and if folder doesn't exist create one
    """
    import matplotlib.pyplot as plt
    import os
    folder_path = file_path.rsplit('/', 1)[0]
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)
    plt.savefig(file_path)




import numpy as np
import pandas as pd

from pyadlml.dataset import DEVICE, TIME, VAL
from pyadlml.dataset.devices import _create_devices
from pyadlml.dataset._dataset import label_data

def create_raw(df_devices, t_res=None, sample_strat='ffill', idle=False):
    dev = df_devices.copy()
    raw = _apply_raw(dev)
    
    if t_res is not None:
        raw = _resample_df(raw, t_res, dev=dev, sample_strat=sample_strat)
        
    return raw

def _apply_raw(df_dev):
    """
        return df:
        | time  | dev_1 | ....  | dev_n |
        --------------------------------
        | ts1   |   1   | ....  |  0    |
    """
    dev_lst = df_dev[DEVICE].unique()

    df = df_dev.pivot(index=TIME, columns=DEVICE, values=VAL)
    df = df.reset_index()

    # set the first element for each device to the opposite value of the 
    # first occurence
    for dev in dev_lst:
        fvi = df[dev].first_valid_index()
        if fvi != 0:
            value = df[dev].iloc[fvi]
            df.loc[0, dev] = not value

    # fill from start to end NaNs with the preceeding correct value
    df = df.ffill().set_index(TIME)
    return df

def _resample_df(df, t_res, dev=None, sample_strat='ffill'):
    resampler = df.resample(t_res, kind='timestamp')
    if sample_strat == 'ffill':
        """
        for an interval the last device value is taken that changed 
        e.g 09:38:29 dev1 <- 0
            09:38:33 dev1 <- 1
            09:38:56 dev1 <- 0 
           09:38:30 | dev1 | 0 
        => 09:39:00 | dev1 | 0
        """
        raw = resampler.ffill()[1:]
    elif sample_strat == 'int_coverage':
        """
        which state has more coverage of the intervall "wins" the assignment
        """
        # first do a forward fill to correclty represent intervalls where no observation falls into
        raw_ff = resampler.ffill()
        
        # then for intervals where multiple sensors trigger choose the most praevalent
        raw_int = resampler.apply(_max_interval, t_res=t_res, dev=dev)
        
        # combine both by filling gaps with the forward fills
        raw = raw_int.where(~raw_int.isnull(), raw_ff)        
        
        #rs_df = self._ffill_bfill_inv_first_dev(rs_df)
    else:
        raise NotImplementedError
        
    return raw


def _max_interval(series: pd.Series, t_res, dev) -> pd.Series:
    """
    cc. kasteren
    
    Parameters
    ----------
        series: pd.Series
            the datapoints falling in the interval for a column
            
        t_res: the resolution of the intervals
            e.g 30s ->  09:38:30, 09:39:00
        devs: representation 3 of devices
            used to identify when values changes
    """
    # return nan if no element matchs, happens till first occurence of data
    if series.empty:
        return np.nan
    
    # if one element matches the interval slot, assign the one
    elif series.size == 1:
        return series
    
    # if there are multiple elements choose the element with the maximal overlap
    else:
        cum_time_0 = pd.Timedelta(seconds=0)
        cum_time_1 = pd.Timedelta(seconds=0)
        
        # deduce the beginning of the interval and the interval
        cur_time = series.index[0].floor(t_res)
        end_time = cur_time + pd.Timedelta(t_res)
        
        # deduce if the value at start of the interval 
        dev_that_changed = dev[dev['time'] == series.index[0]]['device'].values[0]
        if dev_that_changed == series.name:
            prae_val = not bool(series[0])    
        else:
            prae_val = bool(series[0])

        # compute cumulative states in each sensor
        for entry in series.iteritems():
            ts = entry[0]
            cur_val = entry[1]
            if prae_val:
                cum_time_1 += ts-cur_time
            else:
                cum_time_0 += ts-cur_time
            if cur_val != prae_val:
                prae_val = cur_val
            cur_time = ts
        
        # compute cum time to remaining end of interval
        if cur_val:
            cum_time_1 += end_time-cur_time
        else:
            cum_time_0 += end_time-cur_time
                       
        assert cum_time_0 + cum_time_1 == pd.Timedelta(t_res)
        
        # return the state that was more present in the interval
        if cum_time_0 > cum_time_1:
            return 0
        else: 
            return 1



import pandas as pd

def hass_db_2_df(db_url, limit=5000000):
    """
    Parameters
    ----------
    db_url : String
        url to the database 
    """
    query = f"""
    SELECT entity_id, state, last_changed
    FROM states
    WHERE
        state NOT IN ('unknown', 'unavailable')
    ORDER BY last_changed DESC
    LIMIT {limit}
    """
    df = pd.read_sql_query(query, db_url)
    return df

def hass_db_2_data(db_url, device_list, start_time=None):
  """ gets a dataframe with devices in devices list from a start_time up 
  Parameters
  ----------
  Returns
  -------
  df : pd.DataFrame
  df2 : pd.DataFrame
      
  """
  device_dict = {device_list[i]:i for i in range(len(device_list))}
  
  df = hass_db_2_df(db_url)
  df = df[df['entity_id'].isin(device_list)]
  df['time'] = pd.to_datetime(df['last_changed'])
  df['val'] = (df['state'] == 'on').astype(int)
  df['device'] = df['entity_id']
  df = df[[ 'time', 'device','val']]
  
  # encode data
  df['device'] = df["device"].replace(device_dict)
  
  # create feature map
  df2 = pd.DataFrame(device_list, columns=['devices'])
  df2 = df2.reset_index()
  df2.columns = ['id', 'devices']
  return df, df2



"""
this file is to bring datasets into specific representations
"""
import pandas as pd
import numpy as np
from pyadlml.dataset.util import print_df
import dask.dataframe as dd
from pyadlml.util import get_parallel, get_npartitions
from pyadlml.dataset import TIME, ACTIVITY, START_TIME, END_TIME

def label_data(df_devices: pd.DataFrame, df_activities: pd.DataFrame, idle=False):
    """
    for each row in the dataframe select the corresponding activity from the
    timestamp append it as column to df_devices
    Parameters
    ----------
    df_devices : pd.DataFrame
        the only constraint is that the there is a column named time or the index named time
        an example can be raw format: 
                                0   ...      13
        Time                        ...
        2008-03-20 00:34:38  False  ...    True
        2008-03-20 00:34:39  False  ...   False
        ...
    idle : bool
        if true this leads to datapoints not falling into a logged activity to be
        labeled as idle

    Returns
    -------
        dataframe df_devices with appended label column
        Name                    0   ...      13 activity
        Time                        ...         
        2008-03-20 00:34:38  False  ...    True idle
        2008-03-20 00:34:39  False  ...   False act1
    """
    df = df_devices.copy()

    # set time as column and not as index
    if df.index.name == TIME:
        df[ACTIVITY] = df.index
        df = df.reset_index()
    else:
        df[ACTIVITY] = df[TIME].copy()
        df = df.reset_index(drop=True)

    if get_parallel():
        #ddf_activities = dd.from_pandas(df_activities, npartitions=get_npartitions())
        # compute with dask in parallel
        df[ACTIVITY] = dd.from_pandas(df[ACTIVITY], npartitions=get_npartitions()).\
                    map_partitions( # apply lambda functions on each partition
                        lambda df: df.apply(
                            _map_timestamp2activity,
                            df_act=df_activities,
                            idle=idle)).\
                    compute(scheduler='processes')
    else:
        df[ACTIVITY] = df[ACTIVITY].apply(
                            _map_timestamp2activity,
                            df_act=df_activities,
                            idle=idle)
    return df



def _map_timestamp2activity(timestamp, df_act, idle):
    """ given a timestamp map the timestamp to an activity in df_act
    :param time:
        timestamp
        2008-02-26 00:39:25
    :param df_act: 

    :return:
        label of the activity
    """

    # select activity intervalls that the timestamp falls into
    mask = (df_act[START_TIME] <= timestamp) & (timestamp <= df_act[END_TIME])
    matches = df_act[mask]
    match_amount = len(matches.index)

    # 1. case no activity interval matched
    if match_amount == 0 and idle:
        return 'idle'
    elif match_amount == 0 and not idle:
        return pd.NaT

    # 2. case single row matches
    elif match_amount  == 1:
        return matches.activity.values[0]
    
    # 3. case multiple rows
    else:
        print()
        print('*'*70)
        print('ts: ', timestamp)
        print('matches: ', matches)
        print('matches_amount: ', match_amount)
        print('overlap of activities. this should be handled when loading activities')
        print('*'*70)
        raise ValueError




from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets._dataset import DataRep
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM

# model names
from scripts.test_model import BHMMTestModel

BHMM = 'bhmm'
BHMMPC = 'bhmm_pc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'

SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHMM
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_CH
TIME_DIFF = MIN30

# model 2
MODEL_CLASS2 = BHSMM

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
MODEL_NAME2 = 'model_' + MODEL_CLASS2 + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF

# todo this
#MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME
MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME
MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"

MODEL_FOLDER_PATH2 = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME2
MODEL_FILE_PATH2 = MODEL_FOLDER_PATH2 + '/' + MODEL_NAME2 +".joblib"

def load_domain_knowledge(file_path):
    with open(file_path, 'r') as file:
        import json
        data = json.load(file)
        act_data = data['activity_data']
        loc_data = data['loc_data']

    import datetime
    for act_data_point in act_data:
        act_data_point['start'] = datetime.datetime.strptime(
            act_data_point['start'],
            '%H:%M:%S').time()
        act_data_point['end'] = datetime.datetime.strptime(
            act_data_point['end'],
            '%H:%M:%S').time()

    return act_data, loc_data

def main():
    ctrl = Controller()
    ctrl.load_dataset_from_file(DATASET_FILE_PATH)

    print()
    # load model
    ctrl.load_model(MODEL_FILE_PATH, 'bhmm')
    ctrl.load_model(MODEL_FILE_PATH2, 'bhsmm')
    state_sel = ['dental_care', 'outside_activity', 'learning']
    #state_sel = ['dental_care', 'enter_home', 'kitchen_social_activity']
    img_file_path = MODEL_FOLDER_PATH + '/' + 'comparision.png'
    ctrl.compare_dur_dists_and_save('bhmm', 'bhsmm', state_sel, img_file_path)


if __name__ == '__main__':
    print('dataset filepath : ', DATASET_FILE_PATH)
    print('model filepath: ', MODEL_FILE_PATH)
    main()
    print('finished')


import sys
import unittest
data_path = "/home/chris/code/adlml/datasets/casas_aruba/corrected_data.csv"
sys.path.append("/home/chris/code/adlml/pyadlml")


import pyadlml.dataset.casas_aruba as ca

class TestCasasAruba(unittest.TestCase):
    def setUp(self):
        self.data = ca.load(data_path) 
        print(self.data)
        

    def test_create_raw(self):
        #tmp = self.data.create_raw()
        pass


import pandas as pd
from pyadlml.dataset import TIME, DEVICE, VAL
from pyadlml.dataset._dataset import label_data


def create_changepoint(df_devices, t_res=None, idle=False):
    dev = df_devices.copy()
    cp = _apply_changepoint(df_devices.copy())
    
    if t_res is not None:
        resampler = cp.resample(t_res, kind='timestamp')
        cp = resampler.apply(_cp_evaluator, dev=dev)
        
    return cp

def _cp_evaluator(series: pd.Series, dev):
    """ for each element return 1 if there was a change in the given interval or 0 otherwise
    Parameters
    ----------
        series: pd.Series
            conatins name of the column the evaluator operates on
            contains the timestamps for that change in set frame and the value 
            that the specified column has at set timestamp.

            is empty if all columns that have no changing entity

        df: pd.DataFrame
            contains device representation 3. Is for determining the states of other
            device/columns for the timestamps

    Returns: 0|1
    """
    if series.empty:
        return 0
    else:
        if series.sum() == 0:
            return 0
        else:
            return 1

def _apply_changepoint(df):
    """
    Parameters
    ----------
        df: pd.DataFrame
        device representation 3
    """
    # create binary vectors and set all changepoints to true
    df = df.pivot(index=TIME, columns=DEVICE, values=VAL)
    df[(df == False)] = True
    
    # fix rest
    df = df.fillna(False)
    df = df.astype(int)
    return df



""" IMPORTANT COPYRIGHT NOTICE:
        the files are hosted on a drive temporarily for debugging only as some datasets require subscription
        and the code for fetching them correctly still has to be implemented.
        TODO add normal procedure to get files from official sources as soon as possible.
"""

import os
from pyadlml.dataset.io import fetch_handler as _fetch_handler

import pyadlml.dataset._datasets.aras as aras
ARAS_URL = 'https://mega.nz/file/hVpRADoZ#GLLZDV4Y-vgdEeEDTXnFxeG3eKllhTljMM1RK-eGyh4'
ARAS_FILENAME = 'aras.zip'

import pyadlml.dataset._datasets.amsterdam as amsterdam
AMSTERDAM_URL = 'https://mega.nz/file/AYhzDLaS#n-CMzBO_raNAgn2Ep1GNgbhah0bHQzuA48PqO_ODEAg'
AMSTERDAM_FILENAME = 'amsterdam.zip'

import pyadlml.dataset._datasets.casas_aruba as casas_aruba
CASAS_ARUBA_URL = 'https://mega.nz/file/QA5hEToD#V0ypxFsxiwWgVV49OzhsX8RnMNTX8MYSUM2TLL1xX6w'
CASAS_ARUBA_FILENAME = 'casas_aruba.zip'

import pyadlml.dataset._datasets.mitlab as mitlab
MITLAB_URL = 'https://mega.nz/file/MB4BFL6S#8MjAQoS-j0Lje1UFoWUMOCay2FcdpVfla6p9MTe4SQM'
MITLAB_FILENAME = 'mitlab.zip'

import pyadlml.dataset._datasets.uci_adl_binary as uci_adl_binary 
from pyadlml.dataset._datasets.uci_adl_binary import fix_OrdonezB_ADLS
UCI_ADL_BINARY_URL = 'https://mega.nz/file/AQIgDQJD#oximAQFjexTKwNP3WYzlPnOGew06YSQ2ef85vvWGN94'
UCI_ADL_BINARY_FILENAME = 'uci_adl_binary.zip'

import pyadlml.dataset._datasets.activity_assistant as act_assist
TUE_2019_URL = 'https://mega.nz/file/sBoCXBrR#Z5paOUwjTo6GWVPxf39ACDs5faPaNbpHah51Q964PBI'
TUE_2019_FILENAME = 'tuebingen_2019.zip'


def fetch_tuebingen_2019(keep_original=True, cache=True):
    dataset_name = 'tuebingen_2019'

    def load_tuebingen_2019(folder_path):
        return act_assist.load(folder_path, subject='M')

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        TUE_2019_FILENAME, TUE_2019_URL, 
                        load_tuebingen_2019)
    return data

def fetch_uci_adl_binary(keep_original=True, cache=True, subject='OrdonezA'):
    assert subject in ['OrdonezA', 'OrdonezB']
    dataset_name = 'uci_adl_binary'

    def load_uci_adl_binary(folder_path):
        sub_dev_file = folder_path + '/{}_Sensors.txt'.format(subject)
        if subject == 'OrdonezB':
            fix_OrdonezB_ADLS(folder_path + '/OrdonezB_ADLs.txt')
            sub_act_file = folder_path + '/{}_ADLs_corr.txt'.format(subject)
        else:
            sub_act_file = folder_path + '/{}_ADLs.txt'.format(subject)

        return uci_adl_binary.load(sub_dev_file, sub_act_file, subject)

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        UCI_ADL_BINARY_FILENAME, UCI_ADL_BINARY_URL, 
                        load_uci_adl_binary, data_postfix=subject)
    return data
    

def fetch_mitlab(keep_original=True, cache=True, subject='subject1'):
    assert subject in ['subject1', 'subject2']
    dataset_name = 'mitlab'

    def load_mitlab(folder_path):
        sub_act = folder_path + '/' + subject + "/Activities.csv"
        sub_dev = folder_path + '/' + subject + "/sensors.csv"
        sub_data = folder_path + '/' + subject + "/activities_data.csv"
        return mitlab.load(sub_dev, sub_act, sub_data)

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        MITLAB_FILENAME, MITLAB_URL, 
                        load_mitlab, data_postfix=subject)
    return data

def fetch_amsterdam(keep_original=True, cache=True):
    dataset_name = 'amsterdam'

    def load_amsterdam(folder_path):
        sensorData = folder_path + "/kasterenSenseData.txt"
        activityData = folder_path + "/kasterenActData.txt"
        return amsterdam.load(sensorData, activityData)

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        AMSTERDAM_FILENAME, AMSTERDAM_URL, 
                        load_amsterdam)
    return data

def fetch_casas_aruba(keep_original=True, cache=True):
    """
    """
    dataset_name = 'casas_aruba'
    def load_casas_aruba(folder_path):
        from pyadlml.dataset._datasets.casas_aruba import _fix_data
        _fix_data(folder_path + "/data")
        return casas_aruba.load(folder_path + '/corrected_data.csv')

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        CASAS_ARUBA_FILENAME, CASAS_ARUBA_URL, 
                        load_casas_aruba)     
    return data


def fetch_aras(keep_original=True, cache=True):
    """ downloads aras dataset into the datahome folder of pyadlml if 
        it wasn't already downloaded and returns data object

    Parameters
    ----------
    keep_original : bool
        Determines whether the original dataset is kept on drive or removed
    cache : bool
        Determines whether the loaded data object is stored on disk or not   
    
    Returns
    -------
    data : Data.obj
    """
    dataset_name = 'aras'

    def load_aras(folder_path):
        return aras.load(folder_path)

    data = _fetch_handler(keep_original, cache, dataset_name, 
                        ARAS_FILENAME, ARAS_URL, load_aras)     
    return data


import sys
import unittest

deviceData = "/home/chris/code/adlml/datasets/kasteren/kasterenSenseData.txt"
activityData = "/home/chris/code/adlml/datasets/kasteren/kasterenActData.txt"
sys.path.append("/home/chris/code/adlml/pyadlml")

import pyadlml.dataset.kasteren as kasteren
import pyadlml.dataset._dataset as ds

class TestKasteren(unittest.TestCase):
    def setUp(self):
        self.data = kasteren.load(deviceData, activityData) 
        

    def test_set_up(self):
        pass

    def test_create_raw(self):
        t_res = None
        idle = False
        raw = ds.create_raw(self.data.df_devices,self.data.df_activities,t_res=t_res,idle=idle)
        
        # TODO asserts
        assert raw.shape[0] == 2638 and raw.shape[1] == 15

        # test filling nans with idle
        idle = True
        t_res = None
        raw = ds.create_raw(self.data.df_devices,self.data.df_activities,t_res=t_res,idle=idle)

        assert not raw['activity'].isnull().values.any()
        assert raw.shape[0] == 2638 and raw.shape[1] == 15

        # test time discretization 
        idle = False
        t_res = '30s'
        raw = ds.create_raw(self.data.df_devices,self.data.df_activities,t_res=t_res,idle=idle)
        print(raw)

    
    def test_raw_corr(self):
        """ tests the correlation between raw activities"""
        pass


import pandas as pd
import numpy as np
from pyadlml.dataset import START_TIME, END_TIME, ACTIVITY

def activities_duration_dist(df_activities, freq='minutes'):
    """
    """
    assert freq in ['minutes', 'hours', 'seconds']
    df = df_activities.copy()

    # integrate the time difference for activities
    diff = 'total_time'
    df[diff] = df[END_TIME] - df[START_TIME]
    df = df[[ACTIVITY, diff]]
    if freq == 'seconds':
        func = lambda x: x.total_seconds()
    elif freq == 'minutes':
        func = lambda x: x.total_seconds()/60
    else:
        func = lambda x: x.total_seconds()/3600
    df[freq] = df[diff].apply(func)
    return df

def activities_durations(df_activities, freq='minutes'):
    """
    returns a dataframe containing statistics about the activities durations
    """
    df = activities_duration_dist(df_activities, freq=freq)
    df = df.groupby('activity').sum()

    # compute fractions of activities to each other
    norm = df[freq].sum()
    #df[freq] = df[freq].apply(lambda x: x/norm)
    return df

def activities_count(df_activities):
    """
    Returns
    -------
    res pd.Dataframe
                   leave house  use toilet  ...  prepare Dinner  get drink
        occurence           33         111  ...              10         19

    """
    df = df_activities.copy()

    # count the occurences of the activity
    df = df.groupby(ACTIVITY).count()
    df = df.drop(columns=[END_TIME])
    df.columns = ['occurence']
    return df


def activities_dist(df_act, t_range='day', n=1000):
    """
    returns an array where for one (day | week) the activities
    are sampled over time 
    """
    return _create_dist(df_act, label=ACTIVITY, t_range=t_range, n=n)

def _create_dist(df, label='activity', t_range='day', n=1000):
    """
        principle: ancestral sampling
            select an interval at random 
            sample uniform value of the interval
        # todo
            make gaussian Distribution centered at interval instead 
            of uniform 
            solves the problem that beyond interval limits there can't be
            any sample
    """
    assert t_range in ['day', 'week']
    df = df.copy()

    res = pd.DataFrame()
    for i in df[label].unique():
        series = _sample_ts(df[df[label] == i], n)
        res[i] = series
    return res

def activities_transitions(df_act):
    """
    returns the transition matrix (a \times a) of activities
    c_ij describes how often an activity i was followed by activity j
    """
    df = df_act.copy()
    df = df[['activity']]
    df['act_after'] = df['activity'].shift(-1)

    #df = df.groupby("activity")['act_after'].value_counts()).unstack(fill_value=0)
    df = pd.crosstab(df["activity"], df['act_after'])
    return df

def _sample_ts(sub_df, n):
    """
    params:
        sub_df: pd.DataFrame
            dataframe with two time intervals 
            | start_time    | end_time  | .... |
            ------------------------------------
            | ts            | ts        | ...

    returns:
        series: pd.Series
            series of sampled timestamps
    """
    assert 'start_time' in sub_df.columns \
        and 'end_time' in sub_df.columns

    xs = np.empty(shape=(n), dtype=object)    
    for i in range(n):
        idx_sample = np.random.randint(0, sub_df.shape[0])
        interval = sub_df.iloc[idx_sample]
        rand_stamp = _gen_rand_time(ts_min=interval.start_time,  ts_max=interval.end_time)
        rand_stamp = '1990-01-01T'+ rand_stamp            
        tmp = pd.to_datetime(rand_stamp)
        xs[i] = tmp
    return pd.Series(xs)     

def _gen_rand_time(ts_min, ts_max):
    from datetime import timezone, datetime

    # convert to unix timestamp
    unx_min = ts_min.replace(tzinfo=timezone.utc).timestamp()
    unx_max = ts_max.replace(tzinfo=timezone.utc).timestamp()

    # sample uniform and convert back to timestamp
    unx_sample = np.random.randint(unx_min, unx_max)
    ts_sample = datetime.utcfromtimestamp(unx_sample).strftime('%H:%M:%S')
    
    return ts_sample


import numpy as np
import matplotlib.pyplot as plt

# Make some fake data.
a = b = np.arange(0, 3, .02)
c = np.exp(a)
d = c[::-1]
x = np.array(['raw', 'last_fired', 'changepoint'])
activity_count = 10

# the order is 1min,
hmm_freq_1min = np.array([0.246, 0.279, 0.157])
hmm_freq_30sec = np.array([0.298, 0.161, 0.251])
hmm_freq_6sec = np.array([0.318, 0.176, 0.394])

hsmm_freq_1min = np.array([0.309, 0.179, 0.132])
hsmm_freq_30sec = np.array([0.248, 0.161, 0.205])
hsmm_freq_6sec = np.array([0.359, 0.176, 0.381])

# do average over hsmm and hmm to cancel out model influence
freq_1min = (hmm_freq_1min + hsmm_freq_1min)/2
freq_30sec = (hmm_freq_30sec + hsmm_freq_30sec)/2
freq_6sec = (hmm_freq_6sec + hsmm_freq_6sec)/2
freq_mean = (freq_1min + freq_30sec + freq_6sec)/3
chance = np.full(x.shape, 1/activity_count)

# Create plots with pre-defined labels.
fig, ax = plt.subplots()
ax.plot(x, freq_1min, 'k-', label='timeslice 1min')
ax.plot(x, freq_30sec, 'k--', label='timeslice 30sec')
ax.plot(x, freq_6sec, 'k-.', label='timeslice 6sec')
ax.plot(x, freq_mean, 'r', label='mean of all ts')
ax.plot(x, chance, 'k:', label='chance')

legend = ax.legend(loc='upper center')
ax.set_ylabel('ts accuracy')
plt.show()


import numpy as np
import pandas as pd
from pyadlml.dataset.activities import _is_activity_overlapping
from pyadlml.dataset.devices import correct_devices
from pyadlml.dataset.obj import Data

def _get_device_map(data_path):
    """
    Parameters
    ----------
    data_path : string
        path to folder where aras dataset resides
    Returns
    -------
    devices : pd.DataFrame
        dataframe with information about devices 
        the index represents the column of the sensor table
    """
    cols = ['Sensor ID', 'Sensor Type', 'Place']
    devices = pd.DataFrame(columns=cols)
    with open (data_path + 'README', 'r') as file:
        lines = file.readlines()
        # create header
        for line in lines[7:27]:
            s = list(filter(None, line.split('\t')))
            s[-1:] = [s[-1:][0][:-1]]
            tmp = dict(zip(cols, s[1:]))
            devices = devices.append(tmp, ignore_index=True)

        file.close()
    return devices


def _get_activity_map(data_path):
    """
    Parameters
    ----------
    data_path : string
        path to folder where aras dataset resides
    Returns
    -------
    devices : pd.DataFrame
        Contains id matching label of activity 
    """    
    cols = ['ID', 'activity']
    activities = pd.DataFrame(columns=cols)

    with open (data_path + 'README', 'r') as file:
        lines = file.readlines()
        # create header
        for line in lines[33:60]:
            s = list(filter(None, line.split('\t')))
            s[-1:] = [s[-1:][0][:-1]]
            tmp = dict(zip(cols, s))
            activities = activities.append(tmp, ignore_index=True)

        file.close()
    activities['ID'] = activities['ID'].astype(int)
    return activities

def _read_data(data_path, activity_map, device_map):
    #DEF
    file_name = 'DAY_{}.txt'
    cols = list(device_map['Sensor ID']) + (['Resident 1', 'Resident 2'])
    res = pd.DataFrame(columns = cols)

    # read in all files
    for i in range(1,31):
        file_path = data_path + file_name.format(i)
        df = pd.read_table(file_path, sep=' ', header=None)
        df.columns = cols
        if i < 10:
            date = '2000-01-0{}'.format(i)
        else:
            date = '2000-01-{}'.format(i)
        df.index = pd.date_range(date, periods=86400, freq='s')

        # delete all rows where nothing changes
        mask = (df.diff(axis=0).sum(axis=1) == 0.0)
        mask.iloc[0] = False
        df = df[~mask]

        # append all dataframes
        res = res.append(df)

    # label activities
    act_dict = dict(zip(activity_map['ID'],activity_map['activity']))
    res['Resident 1'] = res['Resident 1'].map(act_dict)
    res['Resident 2'] = res['Resident 2'].map(act_dict)
    return res


def _create_activity_df(df, res_name):
    # bring data into representation 3 
    df = pd.DataFrame(df[res_name]).reset_index()
    
    mask_start = df[res_name] != df[res_name].shift(1)
    df_start = df[mask_start]
    df_start = df_start.rename(columns={'index' : 'start_time'})
    df_start = df_start.reset_index(drop=True)

    mask_end = df[res_name] != df[res_name].shift(-1)
    df_end = df[mask_end].reset_index(drop=True)
    
    df_start['end_time'] = df_end['index']
    
    df = df_start[['start_time', 'end_time', res_name]]
    df = df.sort_values(by='start_time')
    df = df.rename(columns={res_name : 'activity'})
    return df


def _create_device_df(df):
    """ gets a raw representation and returns devices in rep3
    Parameters
    ----------
    df : pd.DataFrame
        raw representation 
    """
    df = df.copy().iloc[:,:-2] # exclude activities
    mask = (df.diff(axis=0).sum(axis=1) == 0.0)
    mask.iloc[0] = False
    df = df[~mask]
    
    # is true where the devices change
    mask_0to1 = (df.diff(axis=0) == 1)
    mask_1to0 = (df.diff(axis=0) == -1)
    
    # for every device append the rows where the device changes state
    res = pd.DataFrame(columns=['time', 'device', 'val'])
    for device in df.columns:
        dev_0to1 = pd.DataFrame(df[device][mask_0to1[device]])
        dev_1to0 = pd.DataFrame(df[device][mask_1to0[device]])
        tmp = pd.concat([dev_0to1, dev_1to0]).reset_index()
        tmp.columns=['time', 'val']
        tmp['device'] = device
        
        res = res.append(tmp)
    
    res = res.sort_values(by='time').reset_index(drop=True)
    return res


def load(data_path):
    device_map = _get_device_map(data_path)
    activity_map = _get_activity_map(data_path)
    df = _read_data(data_path, activity_map, device_map)

    df_res1_act = _create_activity_df(df, 'Resident 1')
    df_res2_act = _create_activity_df(df, 'Resident 2')

    assert not _is_activity_overlapping(df_res1_act) \
        or not _is_activity_overlapping(df_res2_act)
    df_dev = _create_device_df(df)

    df_dev = correct_devices(df_dev)

    data = Data(df_res1_act, df_dev)

    data.df_dev_map = device_map
    data.df_act_map = activity_map        
    data.df_activities_res2 = df_res2_act
    return data


import numpy as np
import pandas as pd
from pyadlml.dataset import START_TIME, END_TIME, TIME, \
    TIME, NAME, VAL, DEVICE
from pyadlml.dataset.util import time2int, timestr_2_timedeltas
from pyadlml.dataset.devices import device_rep1_2_rep2, _create_devices, \
                                    _is_dev_rep1, _is_dev_rep2

from pyadlml.dataset.util import timestr_2_timedeltas
from pyadlml.dataset._representations.raw import create_raw
from pyadlml.util import get_npartitions
from dask import delayed
import dask.dataframe as dd

def duration_correlation_parallel(df_dev):
    """ compute the crosscorelation by comparing for every interval the binary values
    between the devices
    
    Parameters
    ----------
        df_dev: pd.DataFrame
            device representation 1 
            columns [time, device, val]
    returns
    -------
        pd.DataFrame (k x k)
        crosscorrelation between each device
    """
    def func(row):
        """ gets two rows and returns a crosstab
        """        
        try:
            td = row.td.to_timedelta64()
        except:
            return None
        states = row.iloc[1:len(row)-1].values.astype(int)
        K = len(states)        
        
        for j in range(K):
            res = np.full((K), 0, dtype='timedelta64[ns]')
            tdiffs = states[j]*states*td            
            row.iloc[1+j] = tdiffs 
        return row
    def create_meta(raw):
        devices = {name : 'object' for name in raw.columns[1:-1]}
        return {**{'time': 'datetime64[ns]', 'td': 'timedelta64[ns]'}, **devices}
        
    dev_lst = df_dev['device'].unique()
    df_dev = df_dev.sort_values(by='time')

    K = len(dev_lst)

    # make off to -1 and on to 1 and then calculate cross correlation between signals
    raw = create_raw(df_dev).applymap(lambda x: 1 if x else -1).reset_index()    
    raw['td'] = raw['time'].shift(-1) - raw['time']
    
    df = dd.from_pandas(raw.copy(), npartitions=get_npartitions())\
                .apply(func, axis=1).drop(columns=['time', 'td']).sum(axis=0)\
                .compute(scheduler='processes')
                #.apply(func, axis=1, meta=create_meta(raw)).drop(columns=['time', 'td']).sum(axis=0)\
    res = pd.DataFrame(data=np.vstack(df.values), columns=df.index, index=df.index)        
    return res/res.iloc[0,0]


def duration_correlation(df_dev):
    """ compute the crosscorelation by comparing for every interval the binary values
    between the devices
    
    Parameters
    ----------
        df_dev: pd.DataFrame
            device representation 1 
            columns [time, device, val]
    returns
    -------
        pd.DataFrame (k x k)
        crosscorrelation between each device
    """
    dev_lst = df_dev['device'].unique()
    df_dev = df_dev.sort_values(by='time')

    K = len(dev_lst)
    crosstab = np.full((K,K), 0, dtype='timedelta64[ns]')

    # make off to -1 and on to 1 and then calculate cross correlation between signals
    
    states = np.full(K, -1)
    prae_time = df_dev.iloc[0].time
    dev_idx = np.where(dev_lst == df_dev.iloc[0].device)[0][0]
    states[dev_idx] = 1
    
    
    # sweep through all 
    i=0
    for row in df_dev.iterrows():
        if i == 0: i+=1; continue

        # for every device determine cross correlation by multiplying
        # the state of the device with the vector of states in order 
        # to know if to subtract or add the time in the previous interval
        nt = row[1].time
        td = (nt - prae_time).to_timedelta64()   
        for j in range(K):
            dev_st = states[j]
            tdiffs = dev_st*states*td
            crosstab[j,:] = crosstab[j,:] + tdiffs

        # update state array with new state and set new time
        dev_idx = np.where(dev_lst == row[1].device)[0][0]
        if row[1].val:
            states[dev_idx] = 1
        else:
            states[dev_idx] = -1
        prae_time = nt
    
    # normalize by the whole time. Diagonal contains full timeframe
    crosstab = crosstab/crosstab[0,0]
    ct = pd.DataFrame(data=crosstab, index=dev_lst, columns=dev_lst)
    return ct


def devices_trigger_count(df):
    """ counts the amount a device was triggered
    Parameters
    ----------
    df : pd.DataFrame
        Contains representation 1 of the devices
        column: [time, device, val]

    Returns
    ------
    """
    assert _is_dev_rep1(df)

    df = df.groupby('device')['device'].count()
    df = pd.DataFrame(df)
    df.columns = ['trigger count']
    return df

def devices_dist(df_dev, t_range='day', n=1000):
    """
    returns an array where for one (day | week) the device
    are sampled over time 
    """
    return _create_dist(df_dev, label=DEVICE, t_range=t_range, n=n)


def trigger_time_diff(df):
    """ adds a column with time in seconds between triggers of devices 
    Parameters
    ----------
    df : pd.DataFrame
        devices in representation 1
    Returns
    -------
    X : np.array
        time deltas in seconds
    """
    # create timediff to the previous trigger
    df['time_diff'] = df['time'].diff()

    # calculate duration to next trigger
    df['row_duration'] = df.time_diff.shift(-1)

    # calculate the time differences for each device sepeartly
    #df = df.sort_values(by=['device','time'])
    #df['time_diff2'] = df['time'].diff()
    #df.loc[df.device != df.device.shift(), 'time_diff2'] = None
    df = df.sort_values(by=['time'])
    df['sec_to_next'] = df['row_duration']/pd.Timedelta(seconds=1)
    X = df['sec_to_next'].values[:-1]

    return X

def devices_on_off_stats(df):
    """
    calculate the time and proportion a device was on vs off
    params:
        df: pd.Dataframe in shape of repr2

    returns: pd.DataFrame
        | device | td_on    | td_off   | frac on   | frac off |
        --------------------------------------------------
        | freezer| 00:10:13 | 27 days   | 0.0025    | 0.999   |
        ....

    """
    # colum names of the final dataframe
    diff = 'diff'
    td_on = 'td_on'
    td_off = 'td_off'
    frac_on = 'frac_on'
    frac_off = 'frac_off'

    df = df.sort_values(START_TIME)

    # calculate total time interval for normalization
    int_start = df.iloc[0,0]
    int_end = df.iloc[df.shape[0]-1,1]
    norm = int_end - int_start

    # calculate time deltas for online time
    df[diff] = df[END_TIME] - df[START_TIME]
    df = df.groupby(DEVICE)[diff].sum()
    df = pd.DataFrame(df)
    df.columns = [td_on]

    df[td_off] = norm - df[td_on]

    # compute percentage
    df[frac_on] = df[td_on].dt.total_seconds() \
                    / norm.total_seconds() 
    df[frac_off] = df[td_off].dt.total_seconds() \
                    / norm.total_seconds()
    return df

def device_tcorr(df, t_windows=['20s']):
    """ computes for every time window the prevalence of device triggers
        for each device

    Parameters
    ----------
    df : pd.DataFrame
        device representation 1
        columns: [time, device, val]

    t_windows : list
        time frames or a single window (string)

    Returns 
    -------
    lst : list of panda dataframes
    """

    t_windows = timestr_2_timedeltas(t_windows)
    
    # create timediff to the previous trigger
    df = df.copy() 
    df['time_diff'] = df['time'].diff()

    #knn
    #    do cumsum for row_duration 
    #    for each row mask the rows that fall into the given area
    dev_list =  df.device.unique()
    
    df.iloc[0,3] = pd.Timedelta(0, 's')
    df['cum_sum'] = df['time_diff'].cumsum()
    
    lst = []
    for t_window in t_windows:
        # create cross table with zeros
        res_df = pd.DataFrame(columns=dev_list, index=dev_list)
        for col in res_df.columns:
            res_df[col].values[:] = 0

        # this whole iterations can be done in parallel
        for row in df.iterrows():
            td = row[1].cum_sum
            dev_name = row[1].device 
            
            df['tmp'] = (td-t_window < df['cum_sum']) & (df['cum_sum'] < td+t_window)
            tmp = df.groupby('device')['tmp'].sum()
            res_df.loc[dev_name] += tmp
        lst.append(res_df)
    return lst

def device_triggers_one_day(df, t_res='1h'):
    """
    computes the amount of triggers of a device for each hour of a day summed
    over all the weeks

    params: df: pd.DataFrame
                repr2 of devices
            t_res: [0,24]h or [0,60]m for a resoltion in hours, minutes
    returns: df
            index: hours
            columsn devices
            values: the amount a device changed states 
    """
    # compute new table
    df['time'] = df['time'].apply(time2int, args=[t_res])
    df = df.groupby(['time', 'device']).sum().unstack()
    df = df.fillna(0)
    df.columns = df.columns.droplevel(0)
    return df




import os
from pyadlml.dataset.io import ENV_DATA_HOME
from pyadlml.util import ENV_PARALLEL

# set data home folder
os.environ[ENV_DATA_HOME] = os.environ['HOME'] + '/pyadlml_data_home'

# default for parallel execution
os.environ[ENV_PARALLEL] = str(False)

import pyadlml.dataset as dataset


import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd

from  pyadlml.dataset.stats.activities import activities_duration_dist, activities_durations,\
    activities_transitions, activities_count, activities_durations, activities_dist
from pyadlml.dataset.activities import add_idle 
from pyadlml.dataset.plot.util import func_formatter_log, func_formatter_min, ridgeline, \
    func_formatter_sec, heatmap, annotate_heatmap, heatmap_square, hm_key_NN

from pyadlml.dataset.util import savefig

ACT_BAR_HM = {  7:(8,4), 9:(9,5), 10:(9,5), 12:(7,5), 22:(10,10), 23:(10,9), 
                26:(10,11)}
ACT_BP_HM = {   7:(8,4), 9:(9,5), 11:(10,6), 22:(10,10), 23: (10,10), 26:(10,11)}
ACT_HM_HM = {   9:(5,5), 10:(6,6), 11:(8,6), 22:(10,10), 26:(11,11)}
ACT_RDG_HM = {  8:(10,8), 11:(10,11), 22:(10,12), 24:(10,12)}

ACT_RDG_YL = {  8:1.1, 11:1.75, 22:3.5, 24:(3.7)}
ACT_RDG_SC = {  8:0.1, 11:0.12, 22:0.16, 24:0.15}


def num_activity_2_figsize(hm, figsize, df):
    if figsize is not None:
        return figsize
    if isinstance(df, pd.DataFrame):
        num_act = len(df['activity'].unique())
    elif isinstance(df, list):
        num_act = len(df)
    else:
        raise ValueError
    return hm_key_NN(hm, num_act)

def hist_counts(df_act=None, df_ac=None, y_scale=None, idle=False, figsize=None, file_path=None):
    """ plots the activities durations against each other
    Parameters
    ----------

    file_path : String
        path where the image will be stored

    Returns
    -------
        Either a figure if file_path is not specified or nothing 
    """
    assert not (df_act is None and df_ac is None)
    assert y_scale in [None, 'log']

    title ='Activity occurrences'
    col_label = 'occurence'
    xlabel = 'counts'
    if df_ac is None:
        df_act = df_act.copy()
        if idle:
            df_act = add_idle(df_act)
        df = activities_count(df_act)
        figsize = num_activity_2_figsize(ACT_BAR_HM, figsize, df_act)
    else: 
        df = df_ac
        figsize = num_activity_2_figsize(ACT_BAR_HM, figsize, list(df['activity']))

    df.reset_index(level=0, inplace=True)
    df = df.sort_values(by=['occurence'], axis=0)
    
    # plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(xlabel)
    ax.barh(df['activity'], df['occurence'])
    if y_scale == 'log':
        ax.set_xscale('log')

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig


def boxplot_duration(df_act, y_scale='norm', idle=False, figsize=None, file_path=None):
    """
        plot a boxplot of activity durations (mean) max min 
    """
    assert y_scale in ['norm', 'log']
    title = 'Activity durations'
    xlabel = 'log seconds'

    if idle:
        df_act = add_idle(df_act)

    figsize = num_activity_2_figsize(ACT_BP_HM, figsize, df_act)
    df = activities_duration_dist(df_act)  
    
    # select data for each device
    activities = df['activity'].unique()
    df['seconds'] = df['minutes']*60 

    dat = []
    for activity in activities:
        df_activity = df[df['activity'] == activity]
        #tmp = np.log(df_device['td'].dt.total_seconds())
        dat.append(df_activity['seconds']) 
    
    # plot boxsplot
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(dat, vert=False)
    ax.set_title(title)
    ax.set_yticklabels(activities, ha='right')
    ax.set_xlabel(xlabel)
    ax.set_xscale('log')

    # create secondary axis with 

    # create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    #ax_top.set_xlabel('time')
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_sec))

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def hist_cum_duration(df_act=None, df_dur=None, y_scale=None, idle=False, figsize=None, file_path=None):
    """ plots the cummulated activities durations in a histogram for each activity 
    """
    assert y_scale in [None, 'log']
    assert not (df_act is None and df_dur is None)

    title = 'Cummulative activity durations'

    if y_scale == 'log':
        xlabel = 'log seconds'
    else: 
        xlabel = 'seconds'

    if df_dur is None:
        if idle:
            df_act = add_idle(df_act.copy())
        act_dur = activities_durations(df_act)
        figsize = num_activity_2_figsize(ACT_BAR_HM, figsize, df_act)
    else:
        act_dur = df_dur
        figsize = num_activity_2_figsize(ACT_BAR_HM, figsize, list(act_dur['activity']))
    df = act_dur[['minutes']]
    df.reset_index(level=0, inplace=True)
    df = df.sort_values(by=['minutes'], axis=0)
    # TODO change in activities duration to return time in seconds
    df['seconds'] = df['minutes']*60 
    
    # plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(xlabel)
    ax.barh(df['activity'], df['seconds'])
    if y_scale == 'log':
        ax.set_xscale('log')
        
        
    # create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.set_xlabel('time')
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_sec))

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def heatmap_transitions(df_act=None, df_trans=None, z_scale=None, figsize=None, \
    idle=False, numbers=True, grid=True, file_path=None):
    """    """
    assert z_scale in [None, 'log'], 'z-scale has to be either of type None or log'
    assert not (df_act is None and df_trans is None)

    title = 'Activity transitions'
    z_label = 'count'

    if df_trans is None:
        df_act = add_idle(df_act) if idle else df_act
        df = activities_transitions(df_act)
    else:
        df = df_trans

    # get the list of cross tabulations per t_window
    act_lst = list(df.columns)
    figsize = num_activity_2_figsize(ACT_HM_HM, figsize, act_lst)
    x_labels = act_lst
    y_labels = act_lst
    values = df.values
    

    log = True if z_scale == 'log' else False
    valfmt = '{x:.0f}'
        
     # begin plotting
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(values, y_labels, x_labels, log=log, ax=ax, cbarlabel=z_label, grid=grid)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"),log=log, valfmt=valfmt)
    ax.set_title(title)
    
    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def ridge_line(df_act=None, act_dist=None, t_range='day', idle=False, \
        n=1000, dist_scale=None, ylim_upper=None, figsize=None, file_path=None):
    """
    Parameters
    ----------
    ylim_upper: float
        height that determines how many ridgelines are displayed. Adjust value to fit all 
        the ridgelines into the plot
    dist_scale: float
        the scale of the distributions of a ridgeline. 
    """
    assert not (df_act is None and act_dist is None)

    title = 'Activity distribution over one day'
    xlabel = 'day'
    if act_dist is None:
        if idle:
            df_act = add_idle(df_act)
        df = activities_dist(df_act.copy(), t_range, n)
    else:
        df = act_dist

    def date_2_second(date):
        """ maps time onto seconds of a day 
        Parameters
        ----------
        date : np.datetime64
            all the dates are on the day 1990-01-01

        Returns
        -------

        """
        val = (date - np.datetime64('1990-01-01')) / np.timedelta64(1, 's')
        total_seconds = 60*60*24
        assert val <= total_seconds and val >= 0
        return int(val)

    
    df = df.apply(np.vectorize(date_2_second))
    # sort every columns values ascending
    for col in df.columns:
        df[col] = df[col].sort_values()

    grouped = [(col, df[col].values) for col in df.columns]

    acts, data = zip(*grouped)
    figsize = num_activity_2_figsize(ACT_RDG_HM, figsize, list(acts))
    if dist_scale is None:
        dist_scale = hm_key_NN(ACT_RDG_SC, len(acts))
    if ylim_upper is None:
        ylim_upper = hm_key_NN(ACT_RDG_YL, len(acts))

    fig, ax = plt.subplots(figsize=figsize)
    ridgeline(data, labels=acts, overlap=.85, fill='tab:blue', n_points=100, dist_scale=dist_scale)
    plt.title(title)

    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.ylim((0, ylim_upper))
    plt.xlabel(xlabel)
    
    # set xaxis labels
    def func(x,p):
        #x = x + 0.5
        #if x == 0.0 or str(x)[-1:] == '5':
        #    return ''
        #else:
        if True:
            if int(x/k) < 10:
                return '0{}:00'.format(int(x/k)+1)
            else:
                return '{}:00'.format(int(x/k)+1)
    a = 0
    b = 60*60*24
    k = (b-a)/24
    
    plt.xlim((a,b))
    tcks_pos = np.arange(0,23)*k + (-0.5 + k)
    
    x_locator = ticker.FixedLocator(tcks_pos)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(func))
    ax.xaxis.set_major_locator(x_locator)
    fig.autofmt_xdate(rotation=45)
    
    plt.grid(zorder=0)

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig


from setuptools import setup, find_packages

with open('README.md', 'r') as readme_file:
  README = readme_file.read()

with open('HISTORY.md', 'r') as history_file:
  HISTORY = history_file.read()

setup_args = dict(
  name = "pyadlml",
  version = "0.0.5.6-alpha",
  url = "https://github.com/tcsvn/pyadlml",
  author = "Christian Meier",
  description = "Sklearn like library supporting numerous Activity of Daily Livings datasets",
  long_description_content_type="text/markdown",
  long_description=README + '\n\n' + HISTORY,
  author_email = "account@meier-lossburg.de",
  license = "MIT",
  packages = find_packages(),
  keywords = ['Activity of Daily Living'],
  classifiers = [
    "Programming Language :: Python :: 3",
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: Visualization"
  ],
)

install_requires = ['numpy', 'pandas', 'joblib', 'mega.py', 'dask[complete]', 
                    'matplotlib', 'scipy', 'sklearn'
]

if __name__ == '__main__':
  setup(**setup_args, install_requires=install_requires)